# -*- coding: utf-8 -*-

import os
import json
import logging
import time
import datetime
from copy import deepcopy

from tornado.httpclient import AsyncHTTPClient, HTTPRequest
import tornado.ioloop
import tornado.web
from tornado import gen

from litepipeline.manager.models.tasks import Tasks
from litepipeline.manager.models.schedules import Schedules
from litepipeline.manager.models.workflows import Workflows
from litepipeline.manager.models.works import Works
from litepipeline.manager.models.services import Services
from litepipeline.manager.utils.app_manager import AppManager
from litepipeline.manager.utils.listener import Connection
from litepipeline.manager.utils.common import Errors, Stage, Status, Event, OperationError, Signal
from litepipeline.manager.config import CONFIG
from litepipeline.manager import logger

LOG = logging.getLogger(__name__)


class Scheduler(object):
    _instance = None
    name = "scheduler"

    def __new__(cls, interval = 1):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.interval = interval
            cls._instance.ioloop_service()
            cls._instance.running_actions = []
            cls._instance.pending_actions = []
            cls._instance.abandoned_actions = []
            cls._instance.tasks = {}
            cls._instance.tasks_stopping = {}
            cls._instance.async_client = AsyncHTTPClient()
            cls._instance.current_select_index = 0
            cls._instance.stop = False
            cls._instance.service_init = True
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def ioloop_service(self):
        self.periodic_schedule = tornado.ioloop.PeriodicCallback(
            self.schedule_service, 
            self.interval * 1000
        )
        self.periodic_schedule.start()
        self.periodic_execute = tornado.ioloop.PeriodicCallback(
            self.execute_service_support_filters, 
            self.interval * 1000
        )
        self.periodic_execute.start()
        self.periodic_crontab = tornado.ioloop.PeriodicCallback(
            self.crontab_service, 
            30 * 1000 # 30 seconds
        )
        self.periodic_crontab.start()
        self.periodic_work = tornado.ioloop.PeriodicCallback(
            self.work_service, 
            self.interval * 1000
        )
        self.periodic_work.start()
        self.periodic_service = tornado.ioloop.PeriodicCallback(
            self.service_service, 
            self.interval * 1000
        )
        self.periodic_service.start()

    @gen.coroutine
    def select_node(self, filters = {}):
        result = None
        try:
            LOG.debug("clients_dict: %s, total_action_slots: %s", Connection.clients_dict, Connection.total_action_slots)
            for node in Connection.clients:
                filters_success = 0
                for f in filters:
                    if f == "node_id":
                        if node.info["node_id"] == filters[f]:
                            filters_success += 1
                        else:
                            break
                    elif f == "node_ip":
                        if node.info["http_host"] == filters[f]:
                            filters_success += 1
                        else:
                            break
                    elif f == "platform":
                        if node.info["platform"].lower() == filters[f].lower():
                            filters_success += 1
                        else:
                            break
                    elif f == "docker_support":
                        if node.info["docker_support"] == filters[f]:
                            filters_success += 1
                        else:
                            break
                if filters_success == len(filters):
                    if "http_host" in node.info and "http_port" in node.info:
                        http_host = node.info["http_host"]
                        http_port = node.info["http_port"]
                        LOG.debug("checking node full, %s:%s", http_host, http_port)
                        url = "http://%s:%s/status/full" % (http_host, http_port)
                        request = HTTPRequest(url = url, method = "GET")
                        try:
                            r = yield self.async_client.fetch(request)
                            if r.code == 200:
                                data = json.loads(r.body.decode("utf-8"))
                                if data["result"] == Errors.OK:
                                    if not data["full"]:
                                        result = node
                                        break
                            else:
                                LOG.error("checking node full failed, %s", url)
                        except ConnectionRefusedError as e:
                            LOG.warning("Scheduler.select_node: GET %s, %s", url, e)
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def select_node_balanced(self):
        result = None
        try:
            LOG.debug("clients_dict: %s, total_action_slots: %s", Connection.clients_dict, Connection.total_action_slots)
            node_ids_tmp = list(Connection.clients_dict.keys())
            node_ids_tmp.sort()
            node_ids = []
            if self.current_select_index >= len(node_ids_tmp):
                self.current_select_index = 0
            if self.current_select_index < 0:
                self.current_select_index = 0
            node_ids.extend(node_ids_tmp[self.current_select_index:])
            node_ids.extend(node_ids_tmp[:self.current_select_index])
            LOG.info("node_ids: %s, %s", node_ids, self.current_select_index)
            for i, node_id in enumerate(node_ids):
                node = Connection.clients_dict[node_id]
                if "http_host" in node.info and "http_port" in node.info:
                    http_host = node.info["http_host"]
                    http_port = node.info["http_port"]
                    LOG.debug("checking node full, %s:%s", http_host, http_port)
                    url = "http://%s:%s/status/full" % (http_host, http_port)
                    request = HTTPRequest(url = url, method = "GET")
                    try:
                        r = yield self.async_client.fetch(request)
                        if r.code == 200:
                            data = json.loads(r.body.decode("utf-8"))
                            if data["result"] == Errors.OK:
                                if not data["full"]:
                                    result = node
                                    self.current_select_index += 1
                                    break
                        else:
                            LOG.error("checking node full failed, %s", url)
                    except ConnectionRefusedError as e:
                        LOG.warning("Scheduler.select_node_balanced: GET %s, %s", url, e)
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def select_executable_action_node(self):
        result_action = None
        result_node = None
        try:
            for action in self.pending_actions:
                if "target_env" in action and action["target_env"]: # execute by specific node
                    node = yield self.select_node(filters = action["target_env"])
                    if node and action["name"] in Event.events:
                        result_action = action
                        break
                    elif node and set(action["condition"]).issubset(set(self.tasks[action["task_id"]]["finished"].keys())):
                        for condition_name in action["condition"]:
                            if "input_data" not in action:
                                action["input_data"] = {condition_name: self.tasks[action["task_id"]]["finished"][condition_name]["result"]["data"]}
                            else:
                                action["input_data"][condition_name] = self.tasks[action["task_id"]]["finished"][condition_name]["result"]["data"]
                        result_action = action
                        result_node = node
                        break
                else:
                    if action["name"] in Event.events:
                        result_action = action
                        break
                    elif set(action["condition"]).issubset(set(self.tasks[action["task_id"]]["finished"].keys())):
                        for condition_name in action["condition"]:
                            if "input_data" not in action:
                                action["input_data"] = {condition_name: self.tasks[action["task_id"]]["finished"][condition_name]["result"]["data"]}
                            else:
                                action["input_data"][condition_name] = self.tasks[action["task_id"]]["finished"][condition_name]["result"]["data"]
                        result_action = action
                        break
        except Exception as e:
            LOG.exception(e)
        raise gen.Return([result_action, result_node])

    @gen.coroutine
    def select_task_node_info(self, task_id, action_name):
        result = {"result": Errors.OK}
        try:
            task_info = Tasks.instance().get(task_id)
            if "result" in task_info and action_name in task_info["result"]:
                action_info = task_info["result"][action_name]
                if "node_id" in action_info and action_info["node_id"]:
                    if "stage" in action_info and action_info["stage"] == Stage.finished:
                        node_id = action_info["node_id"]
                        if node_id in Connection.clients_dict:
                            node = Connection.clients_dict[node_id]
                            result["node_info"] = node.info
                            result["task_info"] = task_info
                        else:
                            Errors.set_result_error("NodeNotExists", result)
                    else:
                        Errors.set_result_error("ActionNotFinished", result)
                else:
                    Errors.set_result_error("ActionNoNodeId", result)
            else:
                Errors.set_result_error("ActionNotRun", result)
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def delete_task_workspace(self, task_ids):
        result = {"result": Errors.OK, "failed_task_ids": []}
        try:
            task_infos = []
            for task_id in task_ids:
                task_info = Tasks.instance().get(task_id)
                task_infos.append({"task_id": task_id, "create_at": task_info["create_at"]})
            for node in Connection.clients:
                node_id = node.info["node_id"]
                http_host = node.info["http_host"]
                http_port = node.info["http_port"]
                LOG.debug("delete task_ids: %s workspace from node: %s(%s:%s)", task_ids, node_id, http_host, http_port)
                url = "http://%s:%s/workspace/delete" % (http_host, http_port)
                request = HTTPRequest(url = url, method = "PUT", body = json.dumps({"task_infos": task_infos}))
                r = yield self.async_client.fetch(request)
                if r.code == 200:
                    result_node = json.loads(r.body.decode("utf-8"))
                    if "failed_task_ids" in result_node:
                        result["failed_task_ids"].extend(result_node["failed_task_ids"])
                    LOG.debug("delete workspace from node: %s(%s:%s), result: %s", node_id, http_host, http_port, result_node)
                else:
                    Errors.set_result_error("OperationFailed", result)
                    LOG.warning("delete workspace from node: %s(%s:%s) failed", node_id, http_host, http_port)
                break
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def pack_task_workspace(self, task_id, action_name, force = False):
        result = {"result": Errors.OK}
        try:
            task_info = Tasks.instance().get(task_id)
            create_at = task_info["create_at"]
            if "result" in task_info and action_name in task_info["result"]:
                action_info = task_info["result"][action_name]
                if "node_id" in action_info and action_info["node_id"]:
                    if "stage" in action_info and action_info["stage"] == Stage.finished:
                        node_id = action_info["node_id"]
                        if node_id in Connection.clients_dict:
                            node = Connection.clients_dict[node_id]
                            http_host = node.info["http_host"]
                            http_port = node.info["http_port"]
                            LOG.debug("pack task_id: %s, action: %s workspace from node: %s(%s:%s)", task_id, action_name, node_id, http_host, http_port)
                            url = "http://%s:%s/workspace/pack" % (http_host, http_port)
                            request = HTTPRequest(url = url, method = "PUT", body = json.dumps({"task_id": task_id, "create_at": create_at, "name": action_name, "force": force}))
                            r = yield self.async_client.fetch(request)
                            if r.code == 200:
                                result = json.loads(r.body.decode("utf-8"))
                                LOG.debug("pack workspace from node: %s(%s:%s), result: %s", node_id, http_host, http_port, result)
                            else:
                                Errors.set_result_error("OperationFailed", result)
                                LOG.warning("pack workspace from node: %s(%s:%s) failed", node_id, http_host, http_port)
                        else:
                            Errors.set_result_error("NodeNotExists", result)
                    else:
                        Errors.set_result_error("ActionNotFinished", result)
                else:
                    Errors.set_result_error("ActionNoNodeId", result)
            else:
                Errors.set_result_error("ActionNotRun", result)
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    def select_executable_action(self):
        result = None
        try:
            for action in self.pending_actions:
                if action["name"] in Event.events:
                    result = action
                    break
                elif set(action["condition"]).issubset(set(self.tasks[action["task_id"]]["finished"].keys())):
                    for condition_name in action["condition"]:
                        if "input_data" not in action:
                            action["input_data"] = {condition_name: self.tasks[action["task_id"]]["finished"][condition_name]["result"]["data"]}
                        else:
                            action["input_data"][condition_name] = self.tasks[action["task_id"]]["finished"][condition_name]["result"]["data"]
                    result = action
                    break
        except Exception as e:
            LOG.exception(e)
        return result

    def select_abandoned_action(self):
        result = None
        try:
            if self.abandoned_actions:
                result = self.abandoned_actions.pop(0)
        except Exception as e:
            LOG.exception(e)
        return result

    def can_load_more_task(self):
        result = False
        try:
            executable_action = self.select_executable_action()
            if executable_action is None and len(self.running_actions) < Connection.get_total_action_slots():
                result = True
            LOG.debug("can_load_more_task, executable_action: %s, running_actions: %s, total_action_slots: %s", executable_action, len(self.running_actions), Connection.total_action_slots)
        except Exception as e:
            LOG.exception(e)
        return result

    def schedule_finish_action(self):
        try:
            for action in self.running_actions:
                pass
        except Exception as e:
            LOG.exception(e)

    def get_running_actions(self, task_id):
        result = []
        try:
            for action in self.running_actions:
                if action["task_id"] == task_id:
                    result.append(action)
        except Exception as e:
            LOG.exception(e)
        return result

    def update_running_action(self, action_data):
        try:
            action_running = None
            now = datetime.datetime.now()
            task_id = action_data["task_id"]
            action_name = action_data["name"]
            for action in self.running_actions:
                if action["task_id"] == task_id and action["name"] == action_name:
                    action["update_at"] = str(now)
                    break
        except Exception as e:
            LOG.exception(e)

    def current_schedule_actions(self):
        result = {"running": [], "pending": []}
        try:
            for action in self.running_actions:
                result["running"].append({
                    "task_id": action["task_id"],
                    "work_id": self.tasks[action["task_id"]]["task_info"]["work_id"],
                    "action_name": action["name"],
                    "update_at": str(action["update_at"]) if "update_at" in action else None,
                })
            for action in self.pending_actions:
                result["pending"].append({
                    "task_id": action["task_id"],
                    "work_id": self.tasks[action["task_id"]]["task_info"]["work_id"],
                    "action_name": action["name"],
                    "update_at": str(action["update_at"]) if "update_at" in action else None,
                })
        except Exception as e:
            LOG.exception(e)
        return result

    def update_finish_action(self, action_result):
        try:
            action_finish = None
            task_id = action_result["task_id"]
            action_name = action_result["name"]
            now = datetime.datetime.now()
            for action in self.running_actions:
                if action["task_id"] == task_id and action["name"] == action_name:
                    action_finish = action
                    break
            if action_finish:
                if action_finish["name"] in Event.events: # event actions
                    task_info = Tasks.instance().get(task_id)
                    task_info["result"][action_finish["name"]] = action_result
                    Tasks.instance().update(task_id, {"result": task_info["result"]})
                    self.running_actions.remove(action_finish)
                else: # normal task actions
                    action_info = {"task_create_at": self.tasks[task_id]["task_info"]["create_at"], "task_name": self.tasks[task_id]["task_info"]["task_name"]}
                    work_id = self.tasks[task_id]["task_info"]["work_id"]
                    service_id = self.tasks[task_id]["task_info"]["service_id"]
                    if action_result["status"] == Status.success:
                        if Stage.stopping in self.tasks[task_id] and self.tasks[task_id][Stage.stopping]: # task is stopping                                
                            self.tasks[task_id]["finished"][action_finish["name"]] = action_result
                            self.running_actions.remove(action_finish)
                            Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.kill, "end_at": now, "result": self.tasks[task_id]["finished"]})
                            if Event.fail in self.tasks[task_id]["event_actions"]:
                                action = self.tasks[task_id]["event_actions"][Event.fail]
                                action["input_data"] = self.tasks[task_id]["task_info"]["input_data"]
                                action["input_data"]["result"] = self.tasks[task_id]["finished"]
                                action_info["action_name"] = action["name"]
                                action["input_data"]["action_info"] = action_info
                                self.pending_actions.append(action)
                            if work_id:
                                work_info = Works.instance().get(work_id)
                                if work_info:
                                    if self.tasks[task_id]["task_info"]["task_name"] in Event.events: # event applications
                                        app = work_info["result"][self.tasks[task_id]["task_info"]["task_name"]]
                                        app["stage"] = Stage.finished
                                        app["status"] = Status.kill
                                        Works.instance().update(work_id, {"result": work_info["result"]})
                                    else: # normal applications
                                        app = work_info["result"][self.tasks[task_id]["task_info"]["task_name"]]
                                        app["stage"] = Stage.finished
                                        app["status"] = Status.kill
                                        Works.instance().update(work_id, {"stage": Stage.finished, "status": Status.kill, "end_at": now, "result": work_info["result"]})
                                        if "event_applications" in work_info["configuration"] and Event.fail in work_info["configuration"]["event_applications"]:
                                            event_app = work_info["configuration"]["event_applications"][Event.fail]
                                            input_data = work_info["input_data"] if work_info["input_data"] else {}
                                            input_data.update(work_info["result"])
                                            new_task_id = Tasks.instance().add(
                                                Event.fail,
                                                event_app["app_id"],
                                                stage = Stage.pending,
                                                input_data = input_data,
                                                work_id = work_info["work_id"]
                                            )
                                            if new_task_id:
                                                work_info["result"][Event.fail] = {
                                                    "name": Event.fail,
                                                    "app_id": event_app["app_id"],
                                                    "task_id": new_task_id,
                                                    "stage": Stage.pending,
                                                    "status": None,
                                                }
                                            else:
                                                raise OperationError("create work[%s]'s task failed" % work_info["work_id"])
                                            Works.instance().update(work_id, {"result": work_info["result"]})
                                else:
                                    LOG.warning("work[%s] not exists", work_id)
                            if service_id:
                                service_info = Services.instance().get(service_id)
                                if service_info:
                                    if service_info["task_id"] == task_id:
                                        Services.instance().update(service_id, {"stage": Stage.finished, "status": Status.kill})
                            del self.tasks[task_id]
                        else:
                            if "actions" in action_result["result"]: # dynamic actions
                                to_actions = {}
                                for action in action_result["result"]["actions"]:
                                    if action["name"] not in self.tasks[task_id]["condition"]:
                                        self.tasks[task_id]["condition"].append(action["name"])
                                    action["task_id"] = task_id
                                    action["app_id"] = self.tasks[task_id]["app_info"]["application_id"]
                                    action["app_sha1"] = self.tasks[task_id]["app_info"]["sha1"]
                                    action["task_create_at"] = self.tasks[task_id]["task_info"]["create_at"]
                                    action["input_data"].update(self.tasks[task_id]["task_info"]["input_data"])
                                    action_info["action_name"] = action["name"]
                                    action["input_data"]["action_info"] = action_info
                                    if "to_action" in action:
                                        if action["to_action"] in to_actions:
                                            to_actions[action["to_action"]].append(action["name"])
                                        else:
                                            to_actions[action["to_action"]] = [action["name"]]
                                    if action["name"] not in self.tasks[task_id]["finished"]:
                                        self.pending_actions.append(action)
                                for action in self.pending_actions:
                                    if action["name"] in to_actions.keys():
                                        action["condition"].extend(to_actions[action["name"]])
                            self.tasks[task_id]["finished"][action_finish["name"]] = action_result
                            self.running_actions.remove(action_finish)
                            finish_condition = self.tasks[task_id]["condition"]
                            current_condition = self.tasks[task_id]["finished"].keys()
                            if len(current_condition) == len(finish_condition): # task finish & success
                                Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.success, "end_at": now, "result": self.tasks[task_id]["finished"]})
                                if Event.success in self.tasks[task_id]["event_actions"]:
                                    action = self.tasks[task_id]["event_actions"][Event.success]
                                    action["input_data"] = self.tasks[task_id]["task_info"]["input_data"]
                                    action["input_data"]["result"] = self.tasks[task_id]["finished"]
                                    action_info["action_name"] = action["name"]
                                    action["input_data"]["action_info"] = action_info
                                    self.pending_actions.append(action)
                                if work_id:
                                    work_info = Works.instance().get(work_id)
                                    if work_info:
                                        if self.tasks[task_id]["task_info"]["task_name"] in Event.events: # event applications
                                            app = work_info["result"][self.tasks[task_id]["task_info"]["task_name"]]
                                            app["stage"] = Stage.finished
                                            app["status"] = Status.success
                                            Works.instance().update(work_id, {"result": work_info["result"]})
                                        else: # normal applications
                                            output_action = self.tasks[task_id]["task_info"]["output_action"]
                                            app = work_info["result"][self.tasks[task_id]["task_info"]["task_name"]]
                                            app["stage"] = Stage.finished
                                            app["status"] = Status.success
                                            app["result"] = self.tasks[task_id]["finished"][output_action]["result"]["data"]

                                            work_current_condition = []
                                            for app_name in work_info["result"]:
                                                app = work_info["result"][app_name]
                                                if "stage" in app and app["stage"] == Stage.finished and app["status"] == Status.success:
                                                    work_current_condition.append(app_name)
                                            if len(work_info["configuration"]["applications"]) == len(work_info["result"]): # work finished
                                                Works.instance().update(work_id, {"stage": Stage.finished, "status": Status.success, "end_at": now, "result": work_info["result"]})
                                                if "event_applications" in work_info["configuration"] and Event.success in work_info["configuration"]["event_applications"]:
                                                    event_app = work_info["configuration"]["event_applications"][Event.success]
                                                    input_data = work_info["input_data"] if work_info["input_data"] else {}
                                                    input_data.update(work_info["result"])
                                                    new_task_id = Tasks.instance().add(
                                                        Event.success,
                                                        event_app["app_id"],
                                                        stage = Stage.pending,
                                                        input_data = input_data,
                                                        work_id = work_info["work_id"]
                                                    )
                                                    if new_task_id:
                                                        work_info["result"][Event.success] = {
                                                            "name": Event.success,
                                                            "app_id": event_app["app_id"],
                                                            "task_id": new_task_id,
                                                            "stage": Stage.pending,
                                                            "status": None,
                                                        }
                                                    else:
                                                        raise OperationError("create work[%s]'s task failed" % work_info["work_id"])
                                                    Works.instance().update(work_id, {"result": work_info["result"]})
                                            else: # work running
                                                for app in work_info["configuration"]["applications"]:
                                                    if app["name"] not in work_info["result"] and set(app["condition"]).issubset(set(work_current_condition)):
                                                        input_data = work_info["input_data"] if work_info["input_data"] else {}
                                                        for app_name in app["condition"]:
                                                            input_data[app_name] = work_info["result"][app_name]["result"]
                                                        new_task_id = Tasks.instance().add(
                                                            app["name"],
                                                            app["app_id"],
                                                            stage = Stage.pending,
                                                            input_data = input_data,
                                                            work_id = work_info["work_id"]
                                                        )
                                                        if new_task_id:
                                                            work_info["result"][app["name"]] = {
                                                                "name": app["name"],
                                                                "app_id": app["app_id"],
                                                                "task_id": new_task_id,
                                                                "stage": Stage.pending,
                                                                "status": None,
                                                            }
                                                        else:
                                                            raise OperationError("create work[%s]'s task failed" % work_info["work_id"])
                                                Works.instance().update(work_info["work_id"], {"result": work_info["result"]})
                                    else:
                                        LOG.warning("work[%s] not exists", work_id)
                                if service_id:
                                    service_info = Services.instance().get(service_id)
                                    if service_info:
                                        if service_info["task_id"] == task_id:
                                            Services.instance().update(service_id, {"stage": Stage.finished, "status": Status.success})
                                del self.tasks[task_id]
                            else: # task running
                                Tasks.instance().update(task_id, {"result": self.tasks[task_id]["finished"]})
                    # action failed
                    else:
                        self.tasks[task_id]["finished"][action_finish["name"]] = action_result
                        pending_actions_tmp = []
                        running_actions_tmp = []
                        for action in self.pending_actions:
                            if action["task_id"] != task_id:
                                pending_actions_tmp.append(action)
                        self.pending_actions = pending_actions_tmp
                        for action in self.running_actions:
                            if action["task_id"] != task_id:
                                running_actions_tmp.append(action)
                            else:
                                self.abandoned_actions.append(action)
                        self.running_actions = running_actions_tmp
                        status = Status.fail # task failed
                        if "signal" in action_finish: # task failed by kill
                            status = Status.kill
                        Tasks.instance().update(task_id, {"stage": Stage.finished, "status": status, "end_at": now, "result": self.tasks[task_id]["finished"]})
                        if Event.fail in self.tasks[task_id]["event_actions"]:
                            action = self.tasks[task_id]["event_actions"][Event.fail]
                            action["input_data"] = self.tasks[task_id]["task_info"]["input_data"]
                            action["input_data"]["result"] = self.tasks[task_id]["finished"]
                            action_info["action_name"] = action["name"]
                            action["input_data"]["action_info"] = action_info
                            self.pending_actions.append(action)
                        if work_id:
                            work_info = Works.instance().get(work_id)
                            if work_info:
                                if self.tasks[task_id]["task_info"]["task_name"] in Event.events: # event applications
                                    app = work_info["result"][self.tasks[task_id]["task_info"]["task_name"]]
                                    app["stage"] = Stage.finished
                                    app["status"] = Status.kill
                                    Works.instance().update(work_id, {"result": work_info["result"]})
                                else: # normal applications
                                    app = work_info["result"][self.tasks[task_id]["task_info"]["task_name"]]
                                    app["stage"] = Stage.finished
                                    app["status"] = status
                                    Works.instance().update(work_id, {"stage": Stage.finished, "status": status, "end_at": now, "result": work_info["result"]})
                                    if "event_applications" in work_info["configuration"] and Event.fail in work_info["configuration"]["event_applications"]:
                                        event_app = work_info["configuration"]["event_applications"][Event.fail]
                                        input_data = work_info["input_data"] if work_info["input_data"] else {}
                                        input_data.update(work_info["result"])
                                        new_task_id = Tasks.instance().add(
                                            Event.fail,
                                            event_app["app_id"],
                                            stage = Stage.pending,
                                            input_data = input_data,
                                            work_id = work_info["work_id"]
                                        )
                                        if new_task_id:
                                            work_info["result"][Event.fail] = {
                                                "name": Event.fail,
                                                "app_id": event_app["app_id"],
                                                "task_id": new_task_id,
                                                "stage": Stage.pending,
                                                "status": None,
                                            }
                                        else:
                                            raise OperationError("create work[%s]'s task failed" % work_info["work_id"])
                                        Works.instance().update(work_id, {"result": work_info["result"]})
                            else:
                                LOG.warning("work[%s] not exists", work_id)
                        if service_id:
                            service_info = Services.instance().get(service_id)
                            if service_info:
                                if service_info["task_id"] == task_id:
                                    Services.instance().update(service_id, {"stage": Stage.finished, "status": status})
                        del self.tasks[task_id]
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def stop_task(self, task_id, signal):
        LOG.info("stop task_id: %s, signal: %s", task_id, signal)
        result = False
        try:
            now = datetime.datetime.now()
            pending_actions_tmp = []
            for action in self.pending_actions:
                if action["task_id"] != task_id:
                    pending_actions_tmp.append(action)
            self.pending_actions = pending_actions_tmp
            signal_actions = []
            for action in self.running_actions:
                if action["task_id"] == task_id:
                    action["signal"] = signal
                    signal_actions.append(action)
            if signal_actions:
                for action in signal_actions:
                    url = "http://%s/action/stop" % action["node"]
                    request = HTTPRequest(url = url, method = "PUT", body = json.dumps(action))
                    r = yield self.async_client.fetch(request)
                    if r.code == 200 and json.loads(r.body.decode("utf-8"))["result"] == Errors.OK:
                        LOG.debug("send signal[%s] to action[%s][%s] success", action["signal"], action["task_id"], action["name"])
                    else:
                        raise OperationError("request node[%s] for stopping action[%s][%s] failed" % (url, action["task_id"], action["name"]))
                self.tasks[task_id][Stage.stopping] = True
                Tasks.instance().update(task_id, {"stage": Stage.stopping})
            else:
                task_name = ""
                work_id = ""
                service_id = ""
                if task_id in self.tasks:
                    work_id = self.tasks[task_id]["task_info"]["work_id"]
                    service_id = self.tasks[task_id]["task_info"]["service_id"]
                    task_name = self.tasks[task_id]["task_info"]["task_name"]
                    Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.kill, "end_at": now, "result": self.tasks[task_id]["finished"]})
                    del self.tasks[task_id]
                else:
                    task_info = Tasks.instance().get(task_id)
                    if task_info:
                        work_id = task_info["work_id"]
                        service_id = task_info["service_id"]
                        task_name = task_info["task_name"]
                    Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.kill, "end_at": now})
                if work_id:
                    work_info = Works.instance().get(work_id)
                    if work_info:
                        app = work_info["result"][task_name]
                        app["stage"] = Stage.finished
                        app["status"] = Status.kill
                        Works.instance().update(work_id, {"stage": Stage.finished, "status": Status.kill, "end_at": now, "result": work_info["result"]})
                    else:
                        LOG.warning("work[%s] not exists", work_id)
                if service_id:
                    service_info = Services.instance().get(service_id)
                    if service_info:
                        if service_info["task_id"] == task_id:
                            Services.instance().update(service_id, {"stage": Stage.finished, "status": Status.kill})
            result = True
        except OperationError as e:
            LOG.error(e)
        except Exception as e:
            LOG.exception(e)
        return result

    @gen.coroutine
    def cancel_action(self, action):
        result = False
        try:
            url = "http://%s/action/cancel" % action["node"]
            request = HTTPRequest(url = url, method = "PUT", body = json.dumps(action))
            r = yield self.async_client.fetch(request)
            if r.code == 200 and json.loads(r.body.decode("utf-8"))["result"] == Errors.OK:
                LOG.debug("send cancel signal to action[%s][%s] success", action["task_id"], action["name"])
            else:
                raise OperationError("request node[%s] for cancelling action[%s][%s] failed" % (url, action["task_id"], action["name"]))
            result = True
        except OperationError as e:
            LOG.error(e)
        except Exception as e:
            LOG.exception(e)
        return result

    def repending_running_actions(self, node_id):
        LOG.warning("repending running actions on node[node_id: %s]", node_id)
        try:
            running_actions_tmp = []
            for action in self.running_actions:
                if action["node_id"] != node_id:
                    running_actions_tmp.append(action)
                else:
                    self.pending_actions.append(action)
            self.running_actions = running_actions_tmp
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def execute_service_support_filters(self): # from RAM pending to RAM running
        LOG.debug("execute_service support filters")
        try:
            action, node = yield self.select_executable_action_node()
            LOG.info("selected action: %s", action)
            if action:
                if not node:
                    node = yield self.select_node_balanced() 
                if node:
                    http_host = node.info["http_host"]
                    http_port = node.info["http_port"]
                    LOG.debug("select node: %s:%s", http_host, http_port)
                    action = self.select_executable_action()
                    LOG.info("seleted action: %s", action)
                    if action:
                        if "input_data" in action:
                            action["input_data"]["ldfs_host"] = CONFIG["ldfs_http_host"]
                            action["input_data"]["ldfs_port"] = CONFIG["ldfs_http_port"]
                            if "action_info" in action["input_data"]:
                                action["input_data"]["action_info"]["http_host"] = http_host
                                action["input_data"]["action_info"]["http_port"] = http_port
                        url = "http://%s:%s/action/run" % (http_host, http_port)
                        request = HTTPRequest(url = url, method = "POST", body = json.dumps(action))
                        r = yield self.async_client.fetch(request)
                        if r.code == 200 and json.loads(r.body.decode("utf-8"))["result"] == Errors.OK:
                            self.pending_actions.remove(action)
                            action["node"] = "%s:%s" % (http_host, http_port)
                            action["node_id"] = node.info["node_id"]
                            self.running_actions.append(action)
                            LOG.debug("migrate action[%s][%s] to running", action["task_id"], action["name"])
                        else:
                            LOG.error("request node[%s] for action[%s][%s] failed", url, action["task_id"], action["name"])
                    else:
                        LOG.debug("no more executable action")
                else:
                    LOG.warning("no selectable node")
            else:
                LOG.debug("no more executable action")
            abandoned_action = self.select_abandoned_action()
            if abandoned_action:
                task_id = abandoned_action["task_id"]
                app_id = abandoned_action["app_id"]
                action_name = abandoned_action["name"]
                success = yield self.cancel_action(abandoned_action)
                if not success:
                    self.abandoned_actions.append(abandoned_action)
                LOG.info("abandon action, app_id: %s, task_id: %s, action: %s, success: %s", app_id, task_id, action_name, success)
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def execute_service(self): # from RAM pending to RAM running
        LOG.debug("execute_service")
        try:
            if not self.stop:
                node = yield self.select_node_balanced()
                if node:
                    http_host = node.info["http_host"]
                    http_port = node.info["http_port"]
                    LOG.debug("select node: %s:%s", http_host, http_port)
                    action = self.select_executable_action()
                    LOG.info("seleted action: %s", action)
                    if action:
                        if "input_data" in action:
                            action["input_data"]["ldfs_host"] = CONFIG["ldfs_http_host"]
                            action["input_data"]["ldfs_port"] = CONFIG["ldfs_http_port"]
                            if "action_info" in action["input_data"]:
                                action["input_data"]["action_info"]["http_host"] = http_host
                                action["input_data"]["action_info"]["http_port"] = http_port
                        url = "http://%s:%s/action/run" % (http_host, http_port)
                        request = HTTPRequest(url = url, method = "POST", body = json.dumps(action))
                        r = yield self.async_client.fetch(request)
                        if r.code == 200 and json.loads(r.body.decode("utf-8"))["result"] == Errors.OK:
                            self.pending_actions.remove(action)
                            action["node"] = "%s:%s" % (http_host, http_port)
                            action["node_id"] = node.info["node_id"]
                            self.running_actions.append(action)
                            LOG.debug("migrate action[%s][%s] to running", action["task_id"], action["name"])
                        else:
                            LOG.error("request node[%s] for action[%s][%s] failed", url, action["task_id"], action["name"])
                    else:
                        LOG.debug("no more executable action")
                else:
                    LOG.warning("no selectable node")
            else:
                LOG.warning("execute_service wait to stop")
            abandoned_action = self.select_abandoned_action()
            if abandoned_action:
                task_id = abandoned_action["task_id"]
                app_id = abandoned_action["app_id"]
                action_name = abandoned_action["name"]
                success = yield self.cancel_action(abandoned_action)
                if not success:
                    self.abandoned_actions.append(abandoned_action)
                LOG.info("abandon action, app_id: %s, task_id: %s, action: %s, success: %s", app_id, task_id, action_name, success)
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def schedule_service(self): # from database pending / recovering to RAM pending
        LOG.debug("schedule_service")
        try:
            if not self.stop:
                load_more_task = self.can_load_more_task()
                LOG.debug("can load more task: %s", load_more_task)
                if load_more_task:
                    task_info = Tasks.instance().get_first()
                    if task_info:
                        task_id = task_info["task_id"]
                        app_id = task_info["application_id"]
                        app_info = AppManager.instance().info(app_id)
                        if app_info:
                            app_config = AppManager.instance().get_app_config(app_id, app_info["sha1"])
                            if app_config:
                                finish_condition = []
                                task_info["output_action"] = app_config["output_action"]
                                if task_info["stage"] == Stage.pending: # load pending task
                                    for action in app_config["actions"]:
                                        action["task_id"] = task_id
                                        action["app_id"] = app_id
                                        action["app_sha1"] = app_info["sha1"]
                                        action["task_create_at"] = task_info["create_at"]
                                        action["input_data"] = deepcopy(task_info["input_data"])
                                        action["input_data"]["action_info"] = {}
                                        action["input_data"]["action_info"]["task_create_at"] = task_info["create_at"]
                                        action["input_data"]["action_info"]["task_name"] = task_info["task_name"]
                                        action["input_data"]["action_info"]["action_name"] = action["name"]
                                        action["docker_registry"] = CONFIG["docker_registry"]
                                        if "target_env" in action["input_data"] and action["input_data"]["target_env"] and action["name"] in action["input_data"]["target_env"]:
                                            if "target_env" not in action:
                                                action["target_env"] = {}
                                            action["target_env"].update(action["input_data"]["target_env"][action["name"]])
                                        finish_condition.append(action["name"])
                                        self.pending_actions.append(action)
                                    event_actions = {}
                                    if "event_actions" in app_config:
                                        for event_name in app_config["event_actions"]:
                                            action = app_config["event_actions"][event_name]
                                            action["name"] = event_name
                                            action["condition"] = []
                                            action["task_id"] = task_id
                                            action["app_id"] = app_id
                                            action["app_sha1"] = app_info["sha1"]
                                            action["task_create_at"] = task_info["create_at"]
                                            action["docker_registry"] = CONFIG["docker_registry"]
                                        event_actions = app_config["event_actions"]
                                    self.tasks[task_id] = {"task_info": task_info, "condition": finish_condition, "app_info": app_info, "finished": {}, "event_actions": event_actions}
                                    Tasks.instance().update(task_id, {"stage": Stage.running, "start_at": datetime.datetime.now()})
                                    work_id = task_info["work_id"]
                                    if work_id:
                                        work_info = Works.instance().get(work_id)
                                        if work_info:
                                            if task_info["task_name"] in work_info["result"]:
                                                work_info["result"][task_info["task_name"]]["stage"] = Stage.running
                                                Works.instance().update(work_id, {"result": work_info["result"]})
                                    service_id = task_info["service_id"]
                                    if service_id:
                                        service_info = Services.instance().get(service_id)
                                        if service_info:
                                            if service_info["task_id"] == task_id:
                                                Services.instance().update(service_id, {"stage": Stage.running})
                                elif task_info["stage"] == Stage.recovering: # load recovering task
                                    actions_tmp = {}
                                    for action in app_config["actions"]: # load configuration actions
                                        action["task_id"] = task_id
                                        action["app_id"] = app_id
                                        action["app_sha1"] = app_info["sha1"]
                                        action["task_create_at"] = task_info["create_at"]
                                        action["input_data"] = deepcopy(task_info["input_data"])
                                        action["input_data"]["action_info"] = {}
                                        action["input_data"]["action_info"]["task_create_at"] = task_info["create_at"]
                                        action["input_data"]["action_info"]["task_name"] = task_info["task_name"]
                                        action["input_data"]["action_info"]["action_name"] = action["name"]
                                        action["docker_registry"] = CONFIG["docker_registry"]
                                        if "target_env" in action["input_data"] and action["input_data"]["target_env"] and action["name"] in action["input_data"]["target_env"]:
                                            if "target_env" not in action:
                                                action["target_env"] = {}
                                            action["target_env"].update(action["input_data"]["target_env"][action["name"]])
                                        finish_condition.append(action["name"])
                                        actions_tmp[action["name"]] = action
                                    event_actions = {}
                                    if "event_actions" in app_config:
                                        for event_name in app_config["event_actions"]:
                                            action = app_config["event_actions"][event_name]
                                            action["name"] = event_name
                                            action["condition"] = []
                                            action["task_id"] = task_id
                                            action["app_id"] = app_id
                                            action["app_sha1"] = app_info["sha1"]
                                            action["task_create_at"] = task_info["create_at"]
                                            action["docker_registry"] = CONFIG["docker_registry"]
                                        event_actions = app_config["event_actions"]
                                    task_result = {}
                                    actions_dynamic = []
                                    for action_name in task_info["result"]: # load result actions
                                        action = task_info["result"][action_name]
                                        if action["stage"] == Stage.finished and action["status"] == Status.success:
                                            if action_name in actions_tmp: # remove success action
                                                del actions_tmp[action_name]
                                            task_result[action_name] = action
                                            if "actions" in action["result"] and action["result"]["actions"]:
                                                actions_dynamic.extend(action["result"]["actions"])
                                    for action in actions_dynamic: # load dynamic actions
                                        to_action = action["to_action"] if "to_action" in action and action["to_action"] else None
                                        if to_action and to_action in actions_tmp:
                                            actions_tmp[to_action]["condition"].append(action["name"])
                                        if action["name"] not in task_result: # failed dynamic action, append into actions_tmp
                                            action["task_id"] = task_id
                                            action["app_id"] = app_id
                                            action["app_sha1"] = app_info["sha1"]
                                            action["task_create_at"] = task_info["create_at"]
                                            action["input_data"]["action_info"] = {}
                                            action["input_data"]["action_info"]["task_create_at"] = task_info["create_at"]
                                            action["input_data"]["action_info"]["task_name"] = task_info["task_name"]
                                            action["input_data"]["action_info"]["action_name"] = action["name"]
                                            action["docker_registry"] = CONFIG["docker_registry"]
                                            if "target_env" in action["input_data"] and action["input_data"]["target_env"] and action["name"] in action["input_data"]["target_env"]:
                                                if "target_env" not in action:
                                                    action["target_env"] = {}
                                                action["target_env"].update(action["input_data"]["target_env"][action["name"]])
                                            if "signal" in action:
                                                del action["signal"]
                                            actions_tmp[action["name"]] = action
                                        finish_condition.append(action["name"])
                                    for name in actions_tmp:
                                        self.pending_actions.append(actions_tmp[name])
                                    self.tasks[task_id] = {"task_info": task_info, "condition": finish_condition, "app_info": app_info, "finished": task_result, "event_actions": event_actions}
                                    LOG.debug("recover task, task_id: %s, condition: %s, finished: %s", task_id, finish_condition, task_result)
                                    Tasks.instance().update(task_id, {"stage": Stage.running, "start_at": datetime.datetime.now()})
                                    work_id = task_info["work_id"]
                                    if work_id:
                                        work_info = Works.instance().get(work_id)
                                        if work_info:
                                            if task_info["task_name"] in work_info["result"]:
                                                work_info["result"][task_info["task_name"]]["stage"] = Stage.running
                                                Works.instance().update(work_id, {"result": work_info["result"]})
                                    service_id = task_info["service_id"]
                                    if service_id:
                                        service_info = Services.instance().get(service_id)
                                        if service_info:
                                            if service_info["task_id"] == task_id:
                                                Services.instance().update(service_id, {"stage": Stage.running})
                                else:
                                    LOG.error("unknown task stage value: %s", task_info["stage"])
                            else:
                                Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.error, "result": {"message": "get app[%s:%s] config file failed" % (app_id, app_info["sha1"])}})
                                LOG.error("Scheduler get app[%s:%s] config file failed", app_id, app_info["sha1"])
                        elif app_info is None:
                            Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.error, "result": {"message": "app[%s] not exists" % app_id}})
                            LOG.error("Scheduler task[%s]'s app_info[%s] not exists", task_id, app_id)
                        else:
                            Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.error, "result": {"message": "get app[%s] failed" % app_id}})
                            LOG.error("Scheduler get task[%s]'s app_info[%s] failed", task_id, app_id)
                    elif task_info is None:
                        LOG.debug("Scheduler no more task to execute")
                    else:
                        LOG.error("Scheduler get task failed")
                else:
                    LOG.warning("no selectable node")
            else:
                LOG.warning("schedule_service wait to stop")
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def crontab_service(self):
        LOG.debug("crontab_service")
        try:
            schedules = Schedules.instance()
            now = datetime.datetime.now()
            for schedule_id in schedules.cache:
                schedule = schedules.cache[schedule_id]
                executed = False if ("executed" not in schedule) or ("executed" in schedule and schedule["executed"] != str(now.date())) else True
                if not executed:
                    if schedule["day_of_month"] == -1 and schedule["day_of_week"] == -1:
                        hour = 0 if schedule["hour"] == -1 else schedule["hour"]
                        minute = 0 if schedule["minute"] == -1 else schedule["minute"]
                        if now.hour == hour and now.minute >= minute and now.minute - minute < 5:
                            if schedule["source"] == Schedules.application:
                                task_id = Tasks.instance().add(schedule["schedule_name"], schedule["source_id"], stage = Stage.pending, input_data = schedule["input_data"])
                                if task_id:
                                    schedule["executed"] = str(now.date())
                                else:
                                    LOG.error("execute schedule every day failed: %s, at: %s, task_id: %s", schedule, now, task_id)
                                LOG.debug("execute schedule every day: %s, at: %s, task_id: %s", schedule, now, task_id)
                            elif schedule["source"] == Schedules.workflow:
                                workflow = Workflows.instance().get(schedule["source_id"])
                                if workflow:
                                    work_id = Works.instance().add(schedule["schedule_name"], schedule["source_id"], stage = Stage.pending, input_data = schedule["input_data"], configuration = workflow["configuration"])
                                    if work_id:
                                        schedule["executed"] = str(now.date())
                                    else:
                                        LOG.error("execute schedule every day failed: %s, at: %s, work_id: %s", schedule, now, work_id)
                                    LOG.debug("execute schedule every day: %s, at: %s, work_id: %s", schedule, now, work_id)
                    elif schedule["day_of_month"] != -1 and schedule["day_of_month"] == now.day:
                        hour = 0 if schedule["hour"] == -1 else schedule["hour"]
                        minute = 0 if schedule["minute"] == -1 else schedule["minute"]
                        if now.hour == hour and now.minute >= minute and now.minute - minute < 5:
                            if schedule["source"] == Schedules.application:
                                task_id = Tasks.instance().add(schedule["schedule_name"], schedule["source_id"], stage = Stage.pending, input_data = schedule["input_data"])
                                if task_id:
                                    schedule["executed"] = str(now.date())
                                else:
                                    LOG.error("execute schedule every month failed: %s, at: %s, task_id: %s", schedule, now, task_id)
                                LOG.debug("execute schedule every month: %s, at: %s, task_id: %s", schedule, now, task_id)
                            elif schedule["source"] == Schedules.workflow:
                                workflow = Workflows.instance().get(schedule["source_id"])
                                if workflow:
                                    work_id = Works.instance().add(schedule["schedule_name"], schedule["source_id"], stage = Stage.pending, input_data = schedule["input_data"], configuration = workflow["configuration"])
                                    if work_id:
                                        schedule["executed"] = str(now.date())
                                    else:
                                        LOG.error("execute schedule every month failed: %s, at: %s, work_id: %s", schedule, now, work_id)
                                    LOG.debug("execute schedule every month: %s, at: %s, work_id: %s", schedule, now, work_id)
                    elif schedule["day_of_week"] != -1 and schedule["day_of_week"] == now.isoweekday():
                        hour = 0 if schedule["hour"] == -1 else schedule["hour"]
                        minute = 0 if schedule["minute"] == -1 else schedule["minute"]
                        if now.hour == hour and now.minute >= minute and now.minute - minute < 5:
                            if schedule["source"] == Schedules.application:
                                task_id = Tasks.instance().add(schedule["schedule_name"], schedule["source_id"], stage = Stage.pending, input_data = schedule["input_data"])
                                if task_id:
                                    schedule["executed"] = str(now.date())
                                else:
                                    LOG.error("execute schedule every week failed: %s, at: %s, task_id: %s", schedule, now, task_id)
                                LOG.debug("execute schedule every week: %s, at: %s, task_id: %s", schedule, now, task_id)
                            elif schedule["source"] == Schedules.workflow:
                                workflow = Workflows.instance().get(schedule["source_id"])
                                if workflow:
                                    work_id = Works.instance().add(schedule["schedule_name"], schedule["source_id"], stage = Stage.pending, input_data = schedule["input_data"], configuration = workflow["configuration"])
                                    if work_id:
                                        schedule["executed"] = str(now.date())
                                    else:
                                        LOG.error("execute schedule every week failed: %s, at: %s, work_id: %s", schedule, now, work_id)
                                    LOG.debug("execute schedule every week: %s, at: %s, work_id: %s", schedule, now, work_id)
                else:
                    LOG.debug("executed schedule: %s", schedule)
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def service_service(self):
        LOG.debug("service_service")
        try:
            services = Services.instance()
            now = datetime.datetime.now()
            if self.stop:
                if len(self.tasks) > 0 and len(self.tasks_stopping) == 0:
                    self.tasks_stopping = deepcopy(self.tasks)
                    for task_id in self.tasks_stopping:
                        yield self.stop_task(task_id, Signal.kill)
            else:
                if self.service_init:
                    LOG.info("init all service")
                    for service_id in services.cache:
                        service = services.cache[service_id]
                        if service["enable"]:
                            task_id = Tasks.instance().add(
                                service["name"],
                                service["application_id"],
                                stage = Stage.pending,
                                input_data = service["input_data"],
                                service_id = service_id
                            )
                            services.update(service_id, {
                                "stage": Stage.pending,
                                "task_id": task_id
                            })
                        else:
                            services.update(service_id, {
                                "stage": Stage.finished,
                                "task_id": "",
                                "status": ""
                            })
                    self.service_init = False
                else:
                    for service_id in services.cache:
                        service = services.cache[service_id]
                        task_id = service["task_id"]
                        if service["enable"] and service["stage"] in ("", Stage.finished):
                            task_id = Tasks.instance().add(
                                service["name"],
                                service["application_id"],
                                stage = Stage.pending,
                                input_data = service["input_data"],
                                service_id = service_id
                            )
                            services.update(service_id, {
                                "stage": Stage.pending,
                                "task_id": task_id,
                                "status": ""
                            })
                        elif not service["enable"] and service["stage"] in (Stage.pending, Stage.recovering, Stage.running) and task_id:
                            yield self.stop_task(task_id, service["signal"])
                        LOG.debug("service: %s", service)
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def work_service(self):
        LOG.debug("work_service")
        try:
            work_info = Works.instance().get_first()
            if work_info:
                if work_info["stage"] == Stage.pending:
                    result = {}
                    for app in work_info["configuration"]["applications"]:
                        if len(app["condition"]) == 0:
                            task_id = Tasks.instance().add(
                                app["name"],
                                app["app_id"],
                                stage = Stage.pending,
                                input_data = work_info["input_data"] if work_info["input_data"] else {},
                                work_id = work_info["work_id"]
                            )
                            if task_id:
                                result[app["name"]] = {
                                    "name": app["name"],
                                    "app_id": app["app_id"],
                                    "task_id": task_id,
                                    "stage": Stage.pending,
                                    "status": None,
                                }
                            else:
                                raise OperationError("create work[%s]'s task failed" % work_info["work_id"])
                elif work_info["stage"] == Stage.recovering:
                    for app_name in work_info["result"]:
                        app = work_info["result"][app_name]
                        if "stage" not in app or ("stage" in app and app["stage"] == Stage.finished and app["status"] in (Status.fail, Status.kill)):
                            app["stage"] = Stage.recovering
                            app["status"] = None
                            task_id = app["task_id"]
                            Tasks.instance().update(task_id, {"stage": Stage.recovering, "status": None})
                    result = work_info["result"]
                Works.instance().update(work_info["work_id"], {"stage": Stage.running, "start_at": datetime.datetime.now(), "result": result})
        except Exception as e:
            LOG.exception(e)

    def close(self):
        try:
            if self.periodic_schedule:
                self.periodic_schedule.stop()
            if self.periodic_crontab:
                self.periodic_crontab.stop()
            if self.periodic_work:
                self.periodic_work.stop()
            if self.periodic_execute:
                self.periodic_execute.stop()
            if self.periodic_service:
                self.periodic_service.stop()
            for task_id in self.tasks_stopping:
                task_info = self.tasks_stopping[task_id]["task_info"]
                if not task_info["service_id"]:
                    Tasks.instance().update(task_id, {"stage": Stage.recovering})
                else:
                    Services.instance().update(task_info["service_id"], {"stage": "", "status": "", "task_id": ""})
            LOG.debug("Scheduler close")
        except Exception as e:
            LOG.exception(e)



# -*- coding: utf-8 -*-

import os
import io
import re
import time
import logging

from tornado import web
from tornado import locale
from tornado import websocket

from litepipeline.manager.config import CONFIG

LOG = logging.getLogger(__name__)


class BaseHandler(web.RequestHandler):
    # def get_current_user(self):
    #     return self.get_secure_cookie("user", max_age_days = 1)

    # def get_current_user_key(self):
    #     return self.get_secure_cookie("user_key", max_age_days = 1)

    # def get_user_locale_value(self):
    #     user_locale = self.get_secure_cookie("user_locale", max_age_days = 1)
    #     if user_locale:
    #         return user_locale
    #     return None

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json')
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header('Access-Control-Allow-Methods', 'GET, POST, PUT, DELETE, OPTIONS')

    def options(self):
        self.set_status(204)
        self.finish()

    def get_user_locale(self):
        pass
        # user_locale = self.get_secure_cookie("user_locale", max_age_days = 1)
        # if user_locale:
        #     return tornado.locale.get(user_locale)
        # return None
    # def __init__(self, application, request, **kwargs):
    #     super(BaseHandler, self).__init__(application, request, kwargs)
    #     self.locale_code = "en_US"
    # pass

    def get_json_argument(self, key, default_value):
        result = default_value
        if key in self.json_data:
            result = self.json_data[key]
        return result

    def get_json_exists_arguments(self, keys):
        result = {}
        for key in keys:
            if key in self.json_data:
                result[key] = self.json_data[key]
        return result


class BaseSocketHandler(websocket.WebSocketHandler):
    def get_current_user(self):
        pass
        # return self.get_secure_cookie("user", max_age_days = 1)

    # def get_current_user_key(self):
    #     return self.get_secure_cookie("user_key", max_age_days = 1)

    # def get_user_locale(self):
    #     user_locale = self.get_secure_cookie("user_locale", max_age_days = 1)
    #     if user_locale:
    #         return tornado.locale.get(user_locale)
    #     return None


@web.stream_request_body
class StreamBaseHandler(BaseHandler):
    PARSE_READY = 0
    PARSE_FILE_PENDING = 1

    def check_xsrf_cookie(self): # ignore xsrf check
        pass

    def prepare(self):
        self.mimetype = self.request.headers.get("Content-Type")
        self.boundary = "--%s" % (self.mimetype[self.mimetype.find("boundary")+9:])
        self.boundary = self.boundary.encode("utf-8")
        self.state = StreamBaseHandler.PARSE_READY
        self.output = None
        self.find_filename = re.compile(b'filename="(.*)"')
        self.find_mimetype = re.compile(b'Content-Type: (.*)')
        self.find_field = re.compile(b'name="(.*)"')
        self.start = time.time()
        self.file_name = ""
        self.file_path = b""
        self.form_arguments = {}

    def data_received(self, data):
        LOG.debug("chunk size: %s", len(data))
        try:
            buff = data.split(self.boundary)
            for index, part in enumerate(buff):
                if part:
                    if part == b"--\r\n":
                        break
                    if self.state == StreamBaseHandler.PARSE_FILE_PENDING:
                        if len(buff) > 1:
                            self.output.write(part[:-2])
                            self.output.close()
                            self.state = StreamBaseHandler.PARSE_READY
                            continue
                        else:
                            self.output.write(part)
                            self.output.flush()
                            continue

                    elif self.state == StreamBaseHandler.PARSE_READY:
                        stream = io.BytesIO(part)
                        stream.readline()
                        form_data_type_line = stream.readline()
                        if form_data_type_line.find(b"filename") > -1:
                            filename = re.search(self.find_filename, form_data_type_line).groups()[0]
                            self.file_name = os.path.split(filename)[-1]
                            self.file_path = os.path.join(CONFIG["data_path"].encode("utf-8"), b"tmp", self.file_name)
                            if self.file_name:
                                self.output = open(self.file_path, "wb")
                                content_type_line = stream.readline()
                                mimetype = re.search(self.find_mimetype, content_type_line).groups()[0]
                                LOG.debug("%s with %s" % (filename, mimetype.strip()))
                                stream.readline()
                                body = stream.read()
                                if len(buff) > index + 1:
                                    self.output.write(body[:-2])
                                    self.output.flush()
                                    self.state = StreamBaseHandler.PARSE_READY
                                else:
                                    self.output.write(body)
                                    self.output.flush()
                                    self.state = StreamBaseHandler.PARSE_FILE_PENDING
                        else:
                            stream.readline()
                            form_name = re.search(self.find_field, form_data_type_line).groups()[0]
                            form_value = stream.readline()
                            if form_name:
                                self.form_arguments[form_name.strip().decode("utf-8")] = form_value.strip().decode("utf-8")
                            self.state = StreamBaseHandler.PARSE_READY
                            LOG.debug("%s = %s" % (form_name.strip(), form_value.strip()))
        except Exception as e:
            LOG.exception(e)

    def get_form_argument(self, key, default_value):
        result = default_value
        try:
            if key in self.form_arguments:
                result = self.form_arguments[key]
        except Exception as e:
            LOG.exception(e)
        return result



# -*- coding: utf-8 -*-

import os
import time
import json
import logging
import shutil
import tarfile

import requests
from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, StreamBaseHandler
from litepipeline.manager.models.workflows import Workflows
from litepipeline.manager.utils.listener import Connection
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateWorkflowHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            name = self.get_json_argument("name", "")
            configuration = self.get_json_argument("configuration", {})
            description = self.get_json_argument("description", "")
            enable = self.get_json_argument("enable", False)
            if name:
                workflow_id = Workflows.instance().add(name, configuration, description = description, enable = enable)
                result["workflow_id"] = workflow_id
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListWorkflowHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            workflow_id = self.get_argument("id", "")
            if workflow_id:
                filters["id"] = workflow_id
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            LOG.debug("ListWorkflowHandler offset: %s, limit: %s, filters: %s", offset, limit, filters)
            r = Workflows.instance().list(offset = offset, limit = limit, filters = filters)
            result["workflows"] = r["workflows"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteWorkflowHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            workflow_id = self.get_argument("workflow_id", "")
            if workflow_id:
                success = Workflows.instance().delete(workflow_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateWorkflowHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            workflow_id = self.get_json_argument("workflow_id", "")
            name = self.get_json_argument("name", "")
            configuration = self.get_json_argument("configuration", {})
            description = self.get_json_argument("description", "")
            enable = self.get_json_argument("enable", None)
            LOG.debug("UpdateWorkflowHandler workflow_id: %s, name: %s, description: %s", workflow_id, name, description)
            if workflow_id and Workflows.instance().get(workflow_id):
                result["workflow_id"] = workflow_id
                data = {}
                if name:
                    data["name"] = name
                if configuration:
                    data["configuration"] = configuration
                if description:
                    data["description"] = description
                if enable is not None:
                    data["enable"] = enable
                if data:
                    success = Workflows.instance().update(workflow_id, data)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class InfoWorkflowHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            workflow_id = self.get_argument("workflow_id", "")
            if workflow_id:
                workflow_info = Workflows.instance().get(workflow_id)
                if workflow_info:
                    result["info"] = workflow_info
                elif workflow_info is None:
                    Errors.set_result_error("AppNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import os
import re
import sys
import json
import time
import urllib
import argparse

import requests
from progress.spinner import Spinner
from litepipeline_helper.models.client import LitePipelineClient

from litepipeline.version import __version__

parser = argparse.ArgumentParser(prog = 'litepipeline')

# common arguments
parser.add_argument("address", help = "manager address, host:port")
parser.add_argument("-W", "--column_width", help = "column max width", type = int, default = 0)
parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
subparsers = parser.add_subparsers(dest = "object", help = 'sub-command help')

# operate with application
parser_app = subparsers.add_parser("app", help = "operate with app API")
subparsers_app = parser_app.add_subparsers(dest = "operation", help = 'sub-command app help')

parser_app_create = subparsers_app.add_parser("create", help = "create application")
parser_app_create.add_argument("-f", "--file", required = True, help = "application's file", default = "")
parser_app_create.add_argument("-n", "--name", required = True, help = "application's name", default = "")
parser_app_create.add_argument("-d", "--description", help = "application's description", default = "")
parser_app_create.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_delete = subparsers_app.add_parser("delete", help = "delete application")
parser_app_delete.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_update = subparsers_app.add_parser("update", help = "update application")
parser_app_update.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_update.add_argument("-f", "--file", help = "application's file", default = "")
parser_app_update.add_argument("-n", "--name", help = "application's name", default = "")
parser_app_update.add_argument("-d", "--description", help = "application's description", default = "")
parser_app_update.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_list = subparsers_app.add_parser("list", help = "list applications")
parser_app_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_app_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_app_list.add_argument("-a", "--app_id", help = "application id filter", default = "")
parser_app_list.add_argument("-n", "--name", help = "application's name filter: '*actions*'", default = "")
parser_app_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_info = subparsers_app.add_parser("info", help = "application's info")
parser_app_info.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_download = subparsers_app.add_parser("download", help = "download application")
parser_app_download.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_download.add_argument("-s", "--sha1", help = "application sha1", default = "")
parser_app_download.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with application history
parser_app_history = subparsers.add_parser("app_history", help = "operate with app_history API")
subparsers_app_history = parser_app_history.add_subparsers(dest = "operation", help = 'sub-command app_history help')

parser_app_history_delete = subparsers_app_history.add_parser("delete", help = "delete application history")
parser_app_history_delete.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_history_delete.add_argument("-H", "--history_id", required = True, help = "history id", default = "")
parser_app_history_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_history_activate = subparsers_app_history.add_parser("activate", help = "activate application history")
parser_app_history_activate.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_history_activate.add_argument("-H", "--history_id", required = True, help = "history id", default = "")
parser_app_history_activate.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_history_list = subparsers_app_history.add_parser("list", help = "list application histories")
parser_app_history_list.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_history_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_app_history_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_app_history_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_history_info = subparsers_app_history.add_parser("info", help = "application history's info")
parser_app_history_info.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_history_info.add_argument("-H", "--history_id", required = True, help = "history id", default = "")
parser_app_history_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with task
parser_task = subparsers.add_parser("task", help = "operate with task API")
subparsers_task = parser_task.add_subparsers(dest = "operation", help = 'sub-command task help')

parser_task_create = subparsers_task.add_parser("create", help = "create task")
parser_task_create.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_task_create.add_argument("-n", "--name", required = True, help = "task's name", default = "")
parser_task_create.add_argument("-i", "--input", help = "task's input data, json string", default = "{}")
parser_task_create.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_delete = subparsers_task.add_parser("delete", help = "delete task")
parser_task_delete.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_task_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_list = subparsers_task.add_parser("list", help = "list tasks")
parser_task_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_task_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_task_list.add_argument("-t", "--task_id", help = "task id filter", default = "")
parser_task_list.add_argument("-a", "--app_id", help = "application id filter", default = "")
parser_task_list.add_argument("-w", "--work_id", help = "work id filter", default = "")
parser_task_list.add_argument("-n", "--name", help = "task's name filter: '*actions*'", default = "")
parser_task_list.add_argument("-s", "--stage", choices = ["pending", "running", "finished"], help = "task's executing stage", default = "")
parser_task_list.add_argument("-S", "--status", choices = ["success", "fail", "kill", "cancel", "terminate", "error"], help = "task's executing status", default = "")
parser_task_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_info = subparsers_task.add_parser("info", help = "task's info")
parser_task_info.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_task_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_rerun = subparsers_task.add_parser("rerun", help = "rerun task")
parser_task_rerun.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_task_rerun.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_recover = subparsers_task.add_parser("recover", help = "recover task")
parser_task_recover.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_task_recover.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_stop = subparsers_task.add_parser("stop", help = "stop task")
parser_task_stop.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_task_stop.add_argument("-g", "--signal", help = "stop task's signal: -9 or -15, default is -9", type = int, default = -9)
parser_task_stop.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with cluster
parser_cluster = subparsers.add_parser("cluster", help = "operate with cluster API")
subparsers_cluster = parser_cluster.add_subparsers(dest = "operation", help = 'sub-command cluster help')

parser_cluster_info = subparsers_cluster.add_parser("info", help = "cluster's info")
parser_cluster_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with workspace
parser_workspace = subparsers.add_parser("workspace", help = "operate with workspace API")
subparsers_workspace = parser_workspace.add_subparsers(dest = "operation", help = 'sub-command workspace help')

parser_workspace_delete = subparsers_workspace.add_parser("delete", help = "delete workspace")
parser_workspace_delete.add_argument("-t", "--task_id", required = True, help = "task id", action = "append")
parser_workspace_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_workspace_download = subparsers_workspace.add_parser("download", help = "download workspace")
parser_workspace_download.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_workspace_download.add_argument("-n", "--name", required = True, help = "action name", default = "")
parser_workspace_download.add_argument("-f", "--force", help = "force repack workspace", action = "store_true")
parser_workspace_download.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with workflow
parser_workflow = subparsers.add_parser("workflow", help = "operate with workflow API")
subparsers_workflow = parser_workflow.add_subparsers(dest = "operation", help = 'sub-command workflow help')

parser_workflow_create = subparsers_workflow.add_parser("create", help = "create workflow")
parser_workflow_create.add_argument("-n", "--name", required = True, help = "workflow's name", default = "")
parser_workflow_create.add_argument("-c", "--config", help = "workflow's json configuration file", default = "")
parser_workflow_create.add_argument("-d", "--description", help = "workflow's description", default = "")
parser_workflow_create.add_argument("-e", "--enable", choices = ["true", "false"], help = "workflow's enable flag", default = "false")
parser_workflow_create.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_workflow_delete = subparsers_workflow.add_parser("delete", help = "delete workflow")
parser_workflow_delete.add_argument("-w", "--workflow_id", required = True, help = "workflow id", default = "")
parser_workflow_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_workflow_update = subparsers_workflow.add_parser("update", help = "update workflow")
parser_workflow_update.add_argument("-w", "--workflow_id", required = True, help = "workflow id", default = "")
parser_workflow_update.add_argument("-n", "--name", help = "workflow's name", default = "")
parser_workflow_update.add_argument("-c", "--config", help = "workflow's json configuration file", default = "")
parser_workflow_update.add_argument("-d", "--description", help = "workflow's description", default = "")
parser_workflow_update.add_argument("-e", "--enable", choices = ["true", "false"], help = "workflow's enable flag", default = "false")
parser_workflow_update.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_workflow_list = subparsers_workflow.add_parser("list", help = "list workflow")
parser_workflow_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_workflow_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_workflow_list.add_argument("-w", "--workflow_id", help = "workflow id filter", default = "")
parser_workflow_list.add_argument("-n", "--name", help = "workflow's name filter: '*actions*'", default = "")
parser_workflow_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_workflow_info = subparsers_workflow.add_parser("info", help = "workflow's info")
parser_workflow_info.add_argument("-w", "--workflow_id", required = True, help = "workflow id", default = "")
parser_workflow_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with work
parser_work = subparsers.add_parser("work", help = "operate with work API")
subparsers_work = parser_work.add_subparsers(dest = "operation", help = 'sub-command work help')

parser_work_create = subparsers_work.add_parser("create", help = "create work")
parser_work_create.add_argument("-w", "--workflow_id", required = True, help = "workflow id", default = "")
parser_work_create.add_argument("-n", "--name", required = True, help = "work's name", default = "")
parser_work_create.add_argument("-i", "--input", help = "work's input data, json string", default = "{}")
parser_work_create.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_work_delete = subparsers_work.add_parser("delete", help = "delete work")
parser_work_delete.add_argument("-w", "--work_id", required = True, help = "work id", default = "")
parser_work_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_work_list = subparsers_work.add_parser("list", help = "list works")
parser_work_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_work_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_work_list.add_argument("-w", "--work_id", help = "work id filter", default = "")
parser_work_list.add_argument("-W", "--workflow_id", help = "workflow id filter", default = "")
parser_work_list.add_argument("-n", "--name", help = "work's name filter: '*actions*'", default = "")
parser_work_list.add_argument("-s", "--stage", choices = ["pending", "running", "finished"], help = "task's executing stage", default = "")
parser_work_list.add_argument("-S", "--status", choices = ["success", "fail", "kill", "cancel", "terminate", "error"], help = "task's executing status", default = "")
parser_work_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_work_info = subparsers_work.add_parser("info", help = "work's info")
parser_work_info.add_argument("-w", "--work_id", required = True, help = "work id", default = "")
parser_work_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_work_rerun = subparsers_work.add_parser("rerun", help = "rerun work")
parser_work_rerun.add_argument("-w", "--work_id", required = True, help = "work id", default = "")
parser_work_rerun.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_work_recover = subparsers_work.add_parser("recover", help = "recover work")
parser_work_recover.add_argument("-w", "--work_id", required = True, help = "work id", default = "")
parser_work_recover.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_work_stop = subparsers_work.add_parser("stop", help = "stop work")
parser_work_stop.add_argument("-w", "--work_id", required = True, help = "work id", default = "")
parser_work_stop.add_argument("-g", "--signal", help = "stop work's signal: -9 or -15, default is -9", type = int, default = -9)
parser_work_stop.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with schedule
parser_schedule = subparsers.add_parser("schedule", help = "operate with schedule API")
subparsers_schedule = parser_schedule.add_subparsers(dest = "operation", help = 'sub-command schedule help')

parser_schedule_create = subparsers_schedule.add_parser("create", help = "create schedule")
parser_schedule_create.add_argument("-a", "--app_id", help = "application id", default = "")
parser_schedule_create.add_argument("-w", "--workflow_id", help = "workflow id", default = "")
parser_schedule_create.add_argument("-n", "--name", required = True, help = "schedule's name", default = "")
parser_schedule_create.add_argument("-i", "--input", help = "task's input data, json string", default = "{}")
parser_schedule_create.add_argument("-m", "--minute", help = "minute, [0, 59]", type = int, default = -1)
parser_schedule_create.add_argument("-H", "--hour", help = "hour, [0, 23]", type = int, default = -1)
parser_schedule_create.add_argument("-d", "--day_of_month", help = "day of month, [1, 31]", type = int, default = -1)
parser_schedule_create.add_argument("-D", "--day_of_week", help = "day of week, [0, 6] (Sunday = 0)", type = int, default = -1)
parser_schedule_create.add_argument("-e", "--enable", choices = ["true", "false"], help = "schedule's enable flag", default = "false")
parser_schedule_create.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_schedule_update = subparsers_schedule.add_parser("update", help = "update schedule")
parser_schedule_update.add_argument("-s", "--schedule_id", required = True, help = "schedule id", default = "")
parser_schedule_update.add_argument("-a", "--app_id", help = "application id")
parser_schedule_update.add_argument("-w", "--workflow_id", help = "workflow id")
parser_schedule_update.add_argument("-n", "--name", help = "schedule's name")
parser_schedule_update.add_argument("-i", "--input", help = "task's input data, json string")
parser_schedule_update.add_argument("-m", "--minute", help = "minute, [0, 59]", type = int)
parser_schedule_update.add_argument("-H", "--hour", help = "hour, [0, 23]", type = int)
parser_schedule_update.add_argument("-d", "--day_of_month", help = "day of month, [1, 31]", type = int)
parser_schedule_update.add_argument("-D", "--day_of_week", help = "day of week, [0, 6] (Sunday = 0)", type = int)
parser_schedule_update.add_argument("-e", "--enable", choices = ["true", "false"], help = "schedule's enable flag")
parser_schedule_update.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_schedule_delete = subparsers_schedule.add_parser("delete", help = "delete schedule")
parser_schedule_delete.add_argument("-s", "--schedule_id", required = True, help = "schedule id", default = "")
parser_schedule_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_schedule_list = subparsers_schedule.add_parser("list", help = "list schedules")
parser_schedule_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_schedule_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_schedule_list.add_argument("-s", "--schedule_id", help = "schedule id filter", default = "")
parser_schedule_list.add_argument("-a", "--app_id", help = "application id filter", default = "")
parser_schedule_list.add_argument("-w", "--workflow_id", help = "workflow id filter", default = "")
parser_schedule_list.add_argument("-n", "--name", help = "schedule's name filter: '*actions*'", default = "")
parser_schedule_list.add_argument("-e", "--enable", choices = ["true", "false"], help = "schedule's enable flag", default = "")
parser_schedule_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_schedule_info = subparsers_schedule.add_parser("info", help = "schedule's info")
parser_schedule_info.add_argument("-s", "--schedule_id", required = True, help = "schedule id", default = "")
parser_schedule_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with service
parser_service = subparsers.add_parser("service", help = "operate with service API")
subparsers_service = parser_service.add_subparsers(dest = "operation", help = 'sub-command service help')

parser_service_create = subparsers_service.add_parser("create", help = "create service")
parser_service_create.add_argument("-n", "--name", required = True, help = "service_id's name", default = "")
parser_service_create.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_service_create.add_argument("-i", "--input", help = "task's input data, json string", default = "{}")
parser_service_create.add_argument("-d", "--description", help = "service's description", default = "")
parser_service_create.add_argument("-e", "--enable", choices = ["true", "false"], help = "service's enable flag", default = "false")
parser_service_create.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_service_update = subparsers_service.add_parser("update", help = "update service")
parser_service_update.add_argument("-s", "--service_id", required = True, help = "service id", default = "")
parser_service_update.add_argument("-a", "--app_id", help = "application id")
parser_service_update.add_argument("-n", "--name", help = "service's name")
parser_service_update.add_argument("-i", "--input", help = "task's input data, json string")
parser_service_update.add_argument("-d", "--description", help = "service's description")
parser_service_update.add_argument("-e", "--enable", choices = ["true", "false"], help = "service's enable flag")
parser_service_update.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_service_delete = subparsers_service.add_parser("delete", help = "delete service")
parser_service_delete.add_argument("-s", "--service_id", required = True, help = "service id", default = "")
parser_service_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_service_list = subparsers_service.add_parser("list", help = "list services")
parser_service_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_service_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_service_list.add_argument("-s", "--service_id", help = "service id filter", default = "")
parser_service_list.add_argument("-a", "--app_id", help = "application id filter", default = "")
parser_service_list.add_argument("-n", "--name", help = "service's name filter: '*service*'", default = "")
parser_service_list.add_argument("-e", "--enable", choices = ["true", "false"], help = "service's enable flag", default = "")
parser_service_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_service_info = subparsers_service.add_parser("info", help = "service's info")
parser_service_info.add_argument("-s", "--service_id", required = True, help = "service_id id", default = "")
parser_service_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

args = parser.parse_args()


def print_table_result(data, fields):
    fields.insert(0, "#")
    field_length_map = {}
    lines = []
    num = 1
    column_max_width = args.column_width
    for field in fields:
        field_length_map[field] = len(field)
    for item in data:
        line = []
        for field in fields:
            if field == "#":
                line.append(str(num))
            else:
                v = str(item[field]) if field in item else ""
                v_len = len(v)
                if column_max_width > 0 and v_len > column_max_width:
                    v_len = column_max_width
                    v = v[:v_len]
                if v_len > field_length_map[field]:
                    field_length_map[field] = v_len
                line.append(v)
        lines.append(tuple(line))
        num += 1
    field_length_map["#"] = len(str(num))
    format_str = ""
    for field in fields:
        field_len = field_length_map[field]
        if field == "#":
            format_str += "%" + " %s" % field_len + "s | "
        else:
            format_str += "%" + "-%s" % field_len + "s | "
    format_str = format_str[:-3]
    print(format_str % tuple(fields))
    for line in lines:
        print(format_str % line)


def main():
    try:
        address = args.address
        object = args.object
        operation = args.operation
        raw = args.raw
        url = "http://%s/%s/%s" % (address, object, operation)
        if address:
            host, port = address.split(":")
            lpl = LitePipelineClient(host, port)
            if object == "app":
                if operation == "list":
                    try:
                        filters = {}
                        if args.name:
                            filters["name"] = args.name
                        if args.app_id:
                            filters["id"] = args.app_id
                        data = lpl.application_list(args.offset, args.limit, filters)
                        if data:
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    data["apps"],
                                    [
                                        "application_id",
                                        "name",
                                        "create_at",
                                        "update_at",
                                    ]
                                )
                    except Exception as e:
                        print(e)
                elif operation == "info":
                    if args.app_id:
                        try:
                            data = lpl.application_info(args.app_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data["app_info"]],
                                        [
                                            "application_id",
                                            "name",
                                            "create_at",
                                            "update_at",
                                            "sha1",
                                            "description",
                                        ]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.app_id:
                        try:
                            data = lpl.application_delete(args.app_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "create":
                    if args.file and args.name:
                        try:
                            data = lpl.application_create(args.file, args.name, args.description)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["app_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "update":
                    if args.app_id:
                        try:
                            data = lpl.application_update(args.app_id, args.file, args.name, args.description)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["app_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        print("error: need app_id(-a, --app_id) parameter")
                elif operation == "download":
                    if args.app_id:
                        try:
                            data = lpl.application_download(args.app_id, sha1 = args.sha1)
                            if data:
                                print("application: %s" % data)
                        except Exception as e:
                            print(e)
                    else:
                        print("error: need app_id(-a, --app_id) parameter")
            elif object == "app_history":
                if operation == "list":
                    try:
                        data = lpl.application_history_list(args.app_id, args.offset, args.limit)
                        if data:
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    data["app_histories"],
                                    [
                                    	"id",
                                        "application_id",
                                        "sha1",
                                        "create_at",
                                    ]
                                )
                    except Exception as e:
                        print(e)
                elif operation == "info":
                    if args.app_id and args.history_id:
                        try:
                            data = lpl.application_history_info(args.app_id, args.history_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data["app_history"]],
                                        [
                                        	"id",
                                            "application_id",
                                            "sha1",
                                            "create_at",
                                            "description",
                                        ]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.app_id and args.history_id:
                        try:
                            data = lpl.application_history_delete(args.app_id, args.history_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "activate":
                    if args.app_id and args.history_id:
                        try:
                            data = lpl.application_history_activate(args.app_id, args.history_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        print("error: need app_id(-a, --app_id) parameter")
            elif object == "task":
                if operation == "list":
                    try:
                        filters = {}
                        if args.task_id:
                            filters["task_id"] = args.task_id
                        if args.app_id:
                            filters["app_id"] = args.app_id
                        if args.work_id:
                            filters["work_id"] = args.work_id
                        if args.name:
                            filters["name"] = args.name
                        if args.stage:
                            filters["stage"] = args.stage
                        if args.status:
                            filters["status"] = args.status
                        data = lpl.task_list(args.offset, args.limit, filters)
                        if data:
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    data["tasks"],
                                    [
                                        "task_id",
                                        "application_id",
                                        "task_name",
                                        "create_at",
                                        "start_at",
                                        "end_at",
                                        "stage",
                                        "status",
                                    ]
                                )
                    except Exception as e:
                        print(e)
                elif operation == "info":
                    if args.task_id:
                        try:
                            data = lpl.task_info(args.task_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data["task_info"]],
                                        [
                                            "task_id",
                                            "application_id",
                                            "task_name",
                                            "create_at",
                                            "start_at",
                                            "end_at",
                                            "stage",
                                            "status",
                                        ]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.task_id:
                        try:
                            data = lpl.task_delete(args.task_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "create":
                    if args.name and args.app_id and args.input:
                        try:
                            data = lpl.task_create(args.name, args.app_id, json.loads(args.input))
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["task_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "rerun":
                    if args.task_id:
                        try:
                            data = lpl.task_rerun(args.task_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "recover":
                    if args.task_id:
                        try:
                            data = lpl.task_recover(args.task_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "stop":
                    if args.task_id and args.signal:
                        try:
                            data = lpl.task_stop(args.task_id, args.signal)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
            elif object == "cluster":
                if operation == "info":
                    try:
                        data = lpl.cluster_info()
                        if data:
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    [data["info"]["manager"]],
                                    [
                                        "http_host",
                                        "http_port",
                                        "tcp_host",
                                        "tcp_port",
                                        "ldfs_http_host",
                                        "ldfs_http_port",
                                        "version",
                                        "data_path",
                                    ]
                                )
                                print()
                                print_table_result(
                                    data["info"]["nodes"],
                                    [
                                        "node_id",
                                        "http_host",
                                        "http_port",
                                        "action_slots",
                                        "version",
                                        "data_path",
                                    ]
                                )
                    except Exception as e:
                        print(e)
            elif object == "workspace":
                if operation == "delete":
                    if args.task_id:
                        try:
                            data = lpl.workspace_delete(args.task_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "download":
                    if args.task_id and args.name:
                        try:
                            spinner = Spinner('Packing ... ')
                            download_ready = lpl.workspace_pack(args.task_id, args.name, force = args.force, callback = spinner.next)
                            if download_ready:
                                print()
                                spinner = Spinner('Downloading ... ')
                                file_path = lpl.workspace_download(args.task_id, args.name, callback = spinner.next)
                                if file_path:
                                    print("\nWorkspace: %s" % file_path)
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
            elif object == "workflow":
                if operation == "list":
                    try:
                        filters = {}
                        if args.name:
                            filters["name"] = args.name
                        if args.workflow_id:
                            filters["id"] = args.workflow_id
                        data = lpl.workflow_list(args.offset, args.limit, filters)
                        if data:
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    data["workflows"],
                                    [
                                        "workflow_id",
                                        "name",
                                        "create_at",
                                        "update_at",
                                        "enable",
                                    ]
                                )
                    except Exception as e:
                        print(e)
                elif operation == "info":
                    if args.workflow_id:
                        try:
                            data = lpl.workflow_info(args.workflow_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data["info"]],
                                        [
                                            "workflow_id",
                                            "name",
                                            "create_at",
                                            "update_at",
                                            "enable",
                                            "description",
                                        ]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.workflow_id:
                        try:
                            data = lpl.workflow_delete(args.workflow_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "create":
                    if args.name:
                        try:
                            configuration = {}
                            enable = True if args.enable == "true" else False
                            if args.config and os.path.exists(args.config) and os.path.isfile(args.config):
                                fp = open(args.config, "r")
                                content = fp.read()
                                fp.close()
                                if content:
                                    configuration = json.loads(content)
                            data = lpl.workflow_create(args.name, configuration, args.description, enable)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["workflow_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "update":
                    if args.workflow_id:
                        try:
                            configuration = {}
                            enable = None
                            if args.enable is not None:
                                enable = True if args.enable == "true" else False
                            if args.config and os.path.exists(args.config) and os.path.isfile(args.config):
                                fp = open(args.config, "r")
                                content = fp.read()
                                fp.close()
                                if content:
                                    configuration = json.loads(content)
                            data = lpl.workflow_update(args.workflow_id, args.name, configuration, args.description, enable)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["workflow_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        print("error: need workflow_id(-w, --workflow_id) parameter")
            elif object == "work":
                if operation == "list":
                    try:
                        filters = {}
                        if args.work_id:
                            filters["work_id"] = args.work_id
                        if args.workflow_id:
                            filters["workflow_id"] = args.workflow_id
                        if args.name:
                            filters["name"] = args.name
                        if args.stage:
                            filters["stage"] = args.stage
                        if args.status:
                            filters["status"] = args.status
                        data = lpl.work_list(args.offset, args.limit, filters)
                        if data:
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    data["works"],
                                    [
                                        "work_id",
                                        "workflow_id",
                                        "name",
                                        "create_at",
                                        "start_at",
                                        "end_at",
                                        "stage",
                                        "status",
                                    ]
                                )
                    except Exception as e:
                        print(e)
                elif operation == "info":
                    if args.work_id:
                        try:
                            data = lpl.work_info(args.work_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data["info"]],
                                        [
                                            "work_id",
                                            "workflow_id",
                                            "name",
                                            "create_at",
                                            "start_at",
                                            "end_at",
                                            "stage",
                                            "status",
                                        ]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.work_id:
                        try:
                            data = lpl.work_delete(args.work_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "create":
                    if args.name and args.workflow_id and args.input:
                        try:
                            data = lpl.work_create(args.name, args.workflow_id, json.loads(args.input))
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["work_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "rerun":
                    if args.work_id:
                        try:
                            data = lpl.work_rerun(args.work_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "recover":
                    if args.work_id:
                        try:
                            data = lpl.work_recover(args.work_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "stop":
                    if args.work_id and args.signal:
                        try:
                            data = lpl.work_stop(args.work_id, args.signal)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
            elif object == "schedule":
                if operation == "list":
                    try:
                        filters = {}
                        if args.schedule_id:
                            filters["schedule_id"] = args.schedule_id
                        if args.app_id:
                            filters["source_id"] = args.app_id
                        if args.workflow_id:
                            filters["source_id"] = args.workflow_id
                        if args.name:
                            filters["name"] = args.name
                        if args.enable:
                            filters["enable"] = args.enable
                        data = lpl.schedule_list(args.offset, args.limit, filters)
                        if data:
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    data["schedules"],
                                    [
                                        "schedule_id",
                                        "source",
                                        "source_id",
                                        "schedule_name",
                                        "create_at",
                                        "update_at",          
                                        "hour",
                                        "minute",
                                        "day_of_month",
                                        "day_of_week",
                                        "enable",
                                    ]
                                )
                    except Exception as e:
                        print(e)
                elif operation == "info":
                    if args.schedule_id:
                        try:
                            data = lpl.schedule_info(args.schedule_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data["schedule_info"]],
                                        [
                                            "schedule_id",
                                            "source",
                                            "source_id",
                                            "schedule_name",
                                            "create_at",
                                            "update_at",
                                            "hour",
                                            "minute",
                                            "day_of_month",
                                            "day_of_week",
                                            "enable",
                                        ]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.schedule_id:
                        try:
                            data = lpl.schedule_delete(args.schedule_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "create":
                    if args.name and args.input and (args.app_id or args.workflow_id):
                        try:
                            enable = True if args.enable == "true" else False
                            if args.app_id:
                                source = "application"
                                source_id = args.app_id
                            elif args.workflow_id:
                                source = "workflow"
                                source_id = args.workflow_id
                            data = lpl.schedule_create(args.name, source, source_id, json.loads(args.input), args.minute, args.hour, args.day_of_month, args.day_of_week, enable)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["schedule_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "update":
                    if args.schedule_id:
                        try:
                            update_data = {}
                            if args.name is not None:
                                update_data["name"] = args.name
                            if args.app_id is not None:
                                update_data["source"] = "application"
                                update_data["source_id"] = args.app_id
                            elif args.workflow_id is not None:
                                update_data["source"] = "workflow"
                                update_data["source_id"] = args.workflow_id
                            if args.input is not None:
                                update_data["input_data"] = args.input
                            if args.minute is not None:
                                update_data["minute"] = args.minute
                            if args.hour is not None:
                                update_data["hour"] = args.hour
                            if args.day_of_month is not None:
                                update_data["day_of_month"] = args.day_of_month
                            if args.day_of_week is not None:
                                update_data["day_of_week"] = args.day_of_week
                            if args.enable is not None:
                                update_data["enable"] = True if args.enable == "true" else False
                            data = lpl.schedule_update(args.schedule_id, update_data)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["schedule_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
            elif object == "service":
                if operation == "list":
                    try:
                        filters = {}
                        if args.service_id:
                            filters["service_id"] = args.service_id
                        if args.app_id:
                            filters["app_id"] = args.app_id
                        if args.name:
                            filters["name"] = args.name
                        if args.enable:
                            filters["enable"] = args.enable
                        data = lpl.service_list(args.offset, args.limit, filters)
                        if data:
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    data["services"],
                                    [
                                        "service_id",
                                        "name",
                                        "application_id",
                                        "task_id",
                                        "create_at",
                                        "update_at",
                                        "stage",
                                        "status",
                                        "enable",
                                    ]
                                )
                    except Exception as e:
                        print(e)
                elif operation == "info":
                    if args.service_id:
                        try:
                            data = lpl.service_info(args.service_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data["service_info"]],
                                        [
                                            "service_id",
                                            "name",
                                            "application_id",
                                            "task_id",
                                            "create_at",
                                            "update_at",
                                            "stage",
                                            "status",
                                            "enable",
                                        ]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.service_id:
                        try:
                            data = lpl.service_delete(args.service_id)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "create":
                    if args.name and args.input and args.app_id:
                        try:
                            enable = True if args.enable == "true" else False
                            data = lpl.service_create(args.name, args.app_id, description = args.description, input_data = json.loads(args.input), enable = enable)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["service_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "update":
                    if args.service_id:
                        try:
                            update_data = {}
                            if args.name is not None:
                                update_data["name"] = args.name
                            if args.app_id is not None:
                                update_data["app_id"] = args.app_id
                            if args.input is not None:
                                update_data["input_data"] = args.input
                            if args.description is not None:
                                update_data["description"] = args.description
                            if args.enable is not None:
                                update_data["enable"] = True if args.enable == "true" else False
                            data = lpl.service_update(args.service_id, update_data)
                            if data:
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["service_id", "result", "message"]
                                    )
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
            else:
                parser.print_help()
        else:
            parser.print_help()
    except Exception as e:
        print(e)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import json
import time
import urllib
import logging

from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.manager.models.services import Services
from litepipeline.manager.utils.app_manager import AppManager
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, Stage, splitall, JSONLoadError
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateServiceHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            name = self.get_json_argument("name", "")
            app_id = self.get_json_argument("app_id", "")
            description = self.get_json_argument("description", "")
            input_data = self.get_json_argument("input_data", {})
            signal = self.get_json_argument("signal", -9)
            enable = True if self.get_json_argument("enable", False) else False
            if AppManager.instance().info(app_id) and name:
                if not isinstance(input_data, dict):
                    raise JSONLoadError("input_data must be dict type")
                service_id = Services.instance().add(
                    name,
                    app_id,
                    description = description,
                    enable = enable,
                    input_data = input_data,
                    signal = signal
                )
                if service_id is not False:
                    result["service_id"] = service_id
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("CreateServiceHandler, name: %s, app_id: %s, enable: %s", name, app_id, enable)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateServiceHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            service_id = self.get_json_argument("service_id", "")
            result["service_id"] = service_id
            data = self.get_json_exists_arguments(
                [
                    "name",
                    "app_id",
                    "description",
                    "input_data",
                    "signal",
                    "enable",
                ]
            )
            if (
                    data and
                    (service_id and Services.instance().get(service_id)) and
                    (("name" in data and data["name"] != "") or "name" not in data) and
                    ("app_id" not in data or ("app_id" in data and AppManager.instance().info(data["app_id"]))) and
                    ("signal" in data and data["signal"] in (-9, -15)) and
                    (("enable" in data and data["enable"] in [True, False]) or "enable" not in data)
                ):
                if "input_data" in data and not isinstance(data["input_data"], dict):
                    raise JSONLoadError("input_data must be dict type")
                if "app_id" in data:
                    data["application_id"] = data["app_id"]
                    del data["app_id"]
                success = Services.instance().update(service_id, data)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("UpdateServiceHandler, service_id: %s, data: %s", service_id, data)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteServiceHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            service_id = self.get_argument("service_id", "")
            if service_id:
                success = Services.instance().delete(service_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class InfoServiceHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            service_id = self.get_argument("service_id", "")
            if service_id:
                service_info = Services.instance().get(service_id)
                if service_info:
                    result["service_info"] = service_info
                    if service_id in Services.instance().cache:
                        result["service_cache_info"] = Services.instance().cache[service_id]
                elif service_info is None:
                    Errors.set_result_error("ServiceNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListServiceHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            service_id = self.get_argument("service_id", "")
            if service_id:
                filters["service_id"] = service_id
            app_id = self.get_argument("app_id", "")
            if app_id:
                filters["app_id"] = app_id
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            enable = self.get_argument("enable", "").lower()
            if enable == "true":
                enable = True
            elif enable == "false":
                enable = False
            else:
                enable = ""
            if enable is not "":
                filters["enable"] = enable
            LOG.debug("ListServiceHandler offset: %s, limit: %s, filters: %s", offset, limit, filters)
            r = Services.instance().list(offset = offset, limit = limit, filters = filters)
            result["services"] = r["services"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()
